from django.views.generic import ListView as BaseListView
from django.views.generic import View

from crudpack.enums import PackViews
from crudpack.services import FieldService

from .base import get_base_view


class ListMixin:
    name = PackViews.LIST

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        pack_info = context.get("pack_info", {})
        object_list = context.get("object_list")
        object_list_count = object_list.count()
        is_paginated = context.get("is_paginated")
        paginator = context.get("paginator")
        page_obj = context.get("page_obj")
        start_index = page_obj.start_index() if is_paginated else 1
        end_index = page_obj.end_index() if is_paginated else object_list_count
        count = paginator.count if is_paginated else object_list_count
        pack_info.update(
            {
                **self.get_fields_data(object_list),
                "start_index": start_index,
                "end_index": end_index,
                "count": count,
            }
        )
        context["pack_info"] = pack_info

        return context

    def get_fields_data(self, queryset):
        service = FieldService(self.pack)
        data = service.get_list_data(queryset)

        return {
            "headers": {
                name: service._get_field_label(name) for name in self.pack.list_fields
            },
            "rows": self.get_rows(data),
        }

    def get_rows(self, data):
        return [
            (instance, {"data": row, "paths": self.pack.get_paths(instance)})
            for instance, row in data
        ]

    def get_paginate_by(self, queryset):
        paginate_by = self.request.GET.get("paginate_by")

        if paginate_by:
            return paginate_by

        return super().get_paginate_by(queryset)


class ListView(View):
    pack = None

    def view(self, request, *args, **kwargs):
        mixins = [ListMixin]
        View = get_base_view(BaseListView, mixins, self.pack)
        View.paginate_by = self.pack.paginate_by
        View.__bases__ = (*self.pack.list_mixins, *View.__bases__)
        view = View.as_view()

        return view(request, *args, **kwargs)

    def dispatch(self, request, *args, **kwargs):
        return self.view(request, *args, **kwargs)



from django.views.generic import DeleteView as BaseDeleteView
from django.views.generic import View

from crudpack.enums import PackViews

from .base import get_base_view


class DeleteMixin:
    name = PackViews.DELETE

    def get_success_url(self):
        self.success_url = self.pack.get_default_success_url()

        return super().get_success_url()


class DeleteView(View):
    pack = None

    def view(self, request, *args, **kwargs):
        mixins = [DeleteMixin]
        View = get_base_view(BaseDeleteView, mixins, self.pack)
        View.__bases__ = (*self.pack.delete_mixins, *View.__bases__)
        view = View.as_view()

        return view(request, *args, **kwargs)

    def dispatch(self, request, *args, **kwargs):
        return self.view(request, *args, **kwargs)



from django.views.generic import CreateView as BaseCreateView
from django.views.generic import View

from crudpack.enums import PackViews
from crudpack.mixins import InlineMixin

from .base import get_base_view


class CreateMixin:
    name = PackViews.CREATE

    def get_success_url(self):
        self.success_url = self.pack.get_default_success_url()

        return super().get_success_url()


class CreateView(View):
    pack = None

    def view(self, request, *args, **kwargs):
        mixins = [CreateMixin]

        if self.pack.inlines:
            mixins.append(InlineMixin)

        View = get_base_view(BaseCreateView, mixins, self.pack)
        View.form_class = self.pack.form_class
        View.fields = self.pack.fields
        View.__bases__ = (*self.pack.form_mixins, *View.__bases__)
        view = View.as_view()

        return view(request, *args, **kwargs)

    def dispatch(self, request, *args, **kwargs):
        return self.view(request, *args, **kwargs)



import inspect
from importlib import import_module


def import_class(module_name, class_name):
    cls = None
    try:
        module = import_module(module_name)
        members = inspect.getmembers(module, inspect.isclass)
        for name, klass in members:
            if name == class_name:
                cls = klass
                break
    except ModuleNotFoundError:
        print("Not found %s" % module_name)
    return cls


def import_mixin(name):
    mixin = import_class("superadmin.mixins", name)
    return mixin


def import_all_mixins():
    mixins = list()
    names = (
        "PermissionRequiredMixin",
        "BreadcrumbMixin",
        "UrlMixin",
        "TemplateMixin",
    )
    for name in names:
        mixin = import_mixin(name)
        if mixin:
            mixins.append(mixin)

    return mixins



from django.utils.module_loading import autodiscover_modules

from crudpack.base import ModelPack
from crudpack.services.packs import packs

__all__ = ["packs", "ModelPack"]


def autodiscover():
    autodiscover_modules("packs")



from django.views.generic import DetailView as BaseDetailView
from django.views.generic import View

from crudpack.enums import PackViews
from crudpack.services import FieldService

from .base import get_base_view


class DetailMixin:
    name = PackViews.DETAIL

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        pack_info = context.get("pack_info", {})
        pack_info.update(self.get_fields_data())
        context["pack_info"] = pack_info

        return context

    def get_fields_data(self):
        service = FieldService(pack=self.pack)
        data, bs_data = service.get_detail_data(self.object)

        return {"fields": data, "bs_fieldsets": bs_data}


class DetailView(View):
    pack = None

    def view(self, request, *args, **kwargs):
        mixins = [DetailMixin]
        View = get_base_view(BaseDetailView, mixins, self.pack)
        View.__bases__ = (*self.pack.detail_mixins, *View.__bases__)
        view = View.as_view()

        return view(request, *args, **kwargs)

    def dispatch(self, request, *args, **kwargs):
        return self.view(request, *args, **kwargs)



from django.core.exceptions import ImproperlyConfigured
from django.db.models import QuerySet
from django.urls import path

from .enums import PackViews
from .views import CreateView, DeleteView, DetailView, ListView, UpdateView

ALL_FIELDS = "__all__"

ALL_VIEWS = {
    PackViews.LIST: ListView,
    PackViews.CREATE: CreateView,
    PackViews.UPDATE: UpdateView,
    PackViews.DETAIL: DetailView,
    PackViews.DELETE: DeleteView,
}


class ModelPack:
    # Views
    model = None
    # Used on Create and Update views
    form_class = None
    # User for passed to Create and Update views for generate forms
    fields = None
    # Used for create ListView with de specified fields
    list_fields = ("__str__",)
    # Used for create DetailView with specified fields
    detail_fields = ()
    # Labels for fields
    default_labels = {}

    allowed_views = tuple(PackViews.__members__.values())
    default_success_url = None

    # Context
    # list_extra_context = {}
    form_extra_context = {}
    detail_extra_context = {}

    # Inlines
    inlines = {}

    # Templates
    list_template_name = None  # Says which list template use
    form_template_name = None  # Says which form template use
    detail_template_name = None  # Says which detail template use
    delete_template_name = None  # Says which delete template use
    _template_names = None

    # Mixins
    list_mixins = ()  # List of mixins that include in ListViews
    form_mixins = ()  # List of mixins that include in Create Views
    update_mixins = ()  # # List of mixins that include in Update Views
    detail_mixins = ()  # List of mixins that include in DetailViews
    delete_mixins = ()  # List of mixins that include in DetailViews

    # Prepopulate
    # slug_field = "slug"
    # prepopulate_slug = ()

    # Options for build queryset
    queryset = None  # Specified custom queryset
    paginate_by = None  # Specified if ListView paginated by

    # Filter and ordering
    # search_fields = ()  # Used for create searchs method by specified fields
    # filter_fields = ()
    # order_by = ()  # User for crate ordering methods by specified fields

    # Urls
    url_list_suffix = "list"
    url_create_suffix = "create"
    url_update_suffix = "update"
    url_detail_suffix = "detail"
    url_delete_suffix = "delete"

    def __init__(self, model, **kwargs):
        if not model:
            error = "The 'model' attribute must be specified."
            raise ImproperlyConfigured(error)

        try:
            self.list_fields = tuple(self.list_fields)
        except TypeError:
            raise ImproperlyConfigured(
                "The 'list_fields' must be an instance of tuple or list"
            )

        if not isinstance(self.detail_fields, (list, dict)):
            try:
                self.detail_fields = tuple(self.detail_fields)
            except TypeError:
                raise ImproperlyConfigured(
                    "The 'detail_fields' must be an instance of tuple, list or dict"
                )

        if not isinstance(self.default_labels, dict):
            raise ImproperlyConfigured(
                "The 'default_labels' must be an instance of dict"
            )

        if not isinstance(self.inlines, dict):
            raise ImproperlyConfigured("The 'inlines' must be an instance of dict")

        self.model = model

        for key, value in kwargs.items():
            setattr(self, key, value)

        if not self.detail_fields:
            self.detail_fields = tuple(field.name for field in self.model._meta.fields)

        self._template_names = {
            PackViews.LIST: self.list_template_name,
            PackViews.CREATE: self.form_template_name,
            PackViews.UPDATE: self.form_template_name,
            PackViews.DETAIL: self.detail_template_name,
            PackViews.DELETE: self.delete_template_name,
        }

        if not isinstance(self.queryset, QuerySet):
            self.queryset = self.model._default_manager.all()

        if not self.allowed_views:
            error = "The 'allowed_views' attribute is required."
            raise ImproperlyConfigured(error)

        if not isinstance(self.allowed_views, (list, tuple)):
            error = "The 'allowed_views' attribute must be a list or tuple."
            raise ImproperlyConfigured(error)

        if not all([isinstance(enum, PackViews) for enum in self.allowed_views]):
            error = "The 'allowed_views' items must be a PackViews element."
            raise ImproperlyConfigured(error)

        if not self.form_class and not self.fields:
            self.fields = ALL_FIELDS

    @property
    def model_info(self):
        return self.model._meta.app_label, self.model._meta.model_name

    def post_save_inlines(self, instance):
        pass

    def get_base_url_name(self, suffix):
        url_suffix = getattr(self, "url_%s_suffix" % suffix)
        base_url_name = "%s_%s_%s" % (*self.model_info, url_suffix)
        return base_url_name

    def get_url_name(self, suffix):
        url_name = "pack:%s" % self.get_base_url_name(suffix)
        return url_name

    def get_urls(self):
        urlpatterns = []

        for enum in self.allowed_views:
            url_name = self.get_base_url_name(enum.value)
            route = "{0}/{1}/".format(enum.param, enum.suffix)
            route = route.replace("//", "/")
            route = route.lstrip("/") if route.startswith("/") else route
            View = ALL_VIEWS.get(enum)
            urlpatterns += [
                path(route=route, view=View.as_view(pack=self), name=url_name)
            ]

        return urlpatterns

    def get_paths(self, instance=None):
        paths = {}

        for enum in self.allowed_views:
            path = "/{0}/{1}/{2}/{3}/".format(
                *self.model_info,
                enum.kwarg,
                enum.suffix,
            )
            path = path.replace("//", "/").replace("//", "/")

            if instance:
                path = path.format(**instance.__dict__)

            paths.update({enum.name.lower(): path})

        return paths

    def get_default_success_url(self):
        if self.default_success_url:
            return self.default_success_url

        name_url = PackViews.LIST.name.lower()
        paths = self.paths
        success_url = paths.get(name_url)

        if not success_url:
            success_url = tuple(paths.values())[0]

        return success_url

    @property
    def urls(self):
        return self.get_urls()

    @property
    def paths(self):
        return self.get_paths()



from django.apps import apps


def register(model):
    """
    @register(Author)
    @register('app.Model')
    class AuthorPack(crudpack.ModelPack):
        pass
    """
    from crudpack import ModelPack, packs

    def _model_pack_wrapper(pack_cls):
        if not model:
            raise ValueError("A model must be passed to be register.")

        model_class = model

        if isinstance(model, str):
            try:
                app_name, model_name = model.split(".")
                model_class = apps.get_model(app_name, model_name)
            except ValueError:
                raise ValueError("Does not exist '%s' model" % model)

        if not issubclass(pack_cls, ModelPack):
            raise ValueError("Wrapped class must subclass ModelPack.")

        packs.register(model_class, pack_cls=pack_cls)

        return pack_cls

    return _model_pack_wrapper



from functools import cached_property

from django.core.exceptions import ImproperlyConfigured
from django.forms import BaseModelForm
from django.forms.models import ModelFormMetaclass as DjangoModelFormMetaclass

from crudpack.services import FieldService


class ModelFormMetaclass(DjangoModelFormMetaclass):
    def __new__(mcs, name, bases, attrs):
        metadata = attrs.get("Meta")
        fieldsets = getattr(metadata, "fieldsets", None)

        if fieldsets:
            field_names = mcs.__fields__(fieldsets)
            fields = getattr(metadata, "fields", ())

            if isinstance(fields, (tuple, list)):
                fields = fields + field_names

            metadata.fields = fields

        new_class = super().__new__(mcs, name, bases, attrs)

        if fieldsets:
            new_class._meta.fieldsets = fieldsets

        return new_class

    def __fields__(field_names):
        if not isinstance(field_names, (tuple, list, dict)):
            raise ImproperlyConfigured(
                "The fieldsets must be an instance of list, tuple or dict"
            )

        fields = FieldService.get_flatten_field_names(field_names)

        return tuple(fields)


class ModelForm(BaseModelForm, metaclass=ModelFormMetaclass):
    def parse(self, field_names):
        def wrap(fields):
            fields = fields if isinstance(fields, (list, tuple)) else (fields,)
            cols = int(12 / len(fields))

            return [(self[field], cols) for field in fields]

        data = list(map(wrap, field_names))

        return data

    @cached_property
    def bs_fieldsets(self):
        field_names = getattr(self._meta, "fieldsets", self.fields.keys())
        fieldsets = (
            field_names.items()
            if isinstance(field_names, dict)
            else [("", field_names)]
        )
        data = [
            {"title": title, "fieldset": self.parse(fieldset)}
            for title, fieldset in fieldsets
        ]

        return data



from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.db.models.base import ModelBase
from django.urls import include, path


class PackService:
    def __init__(self, name="packs"):
        self._registry = {}
        self.name = name

    def register(self, model, pack_cls):
        if isinstance(model, str):
            try:
                app_label, model_name = model.split(".")
                model = apps.get_model(app_label, model_name)
            except ValueError:
                raise ImproperlyConfigured(
                    "The model_name passed"
                    "must be contains app label and model name"
                    "like 'app_label.model_name'"
                )

        if not isinstance(model, ModelBase):
            raise ImproperlyConfigured(
                "The model passed cannot be registered.",
            )

        if model._meta.abstract:
            raise ImproperlyConfigured(
                "The abstract model %s cannot be registered." % model.__name__
            )

        if model in self._registry:
            raise ImproperlyConfigured(
                "The model %s is already registered" % model.__name__
            )

        self._registry[model] = pack_cls(model)

    def is_registered(self, model):
        return model in self._registry

    def get_default_urls(self):
        urlpatterns = []

        for pack in self._registry.values():
            urlpatterns += [
                path(
                    "%s/%s/" % pack.model_info,
                    include(pack.urls),
                ),
            ]

        return urlpatterns

    @property
    def urls(self):
        return self.get_default_urls(), "packs", self.name


packs = PackService()



from django.shortcuts import redirect


class Inline:
    instance = None

    def __init__(self, formset):
        self.formset = formset
        self.headers = [field.label for field in formset.form.base_fields.values()]

    def is_valid(self):
        return self.formset.is_valid()

    def save(self):
        self.formset.instance = self.instance
        self.formset.save()

    @property
    def errors(self):
        return self.formset.errors


class InlineMixin:
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        inlines = self.get_inlines()

        if inlines:
            pack_info = context.get("pack_info", {})
            pack_info.update({"inlines": inlines})
            context["pack_info"] = pack_info

        return context

    def get_inlines(self):
        kwargs = self.get_form_kwargs()

        return {
            name: Inline(inline_class(**kwargs))
            for name, inline_class in self.pack.inlines.items()
        }

    def form_valid(self, form):
        inlines = self.get_inlines().values()

        if any([not inline.is_valid() for inline in inlines]):
            form.inlines = inlines
            return self.form_invalid(form)

        self.object = form.save()

        for inline in inlines:
            inline.instance = self.object
            inline.save()

        self.pack.post_save_inlines(self.object)

        return redirect(self.get_success_url())

    def form_invalid(self, form):
        inlines = getattr(form, "inlines", [])

        for inline in inlines:
            for error in inline.errors:
                form.errors.update(error)

        return super().form_invalid(form)



from .create import CreateView
from .delete import DeleteView
from .detail import DetailView
from .list import ListView
from .update import UpdateView

__all__ = [
    "CreateView",
    "DeleteView",
    "DetailView",
    "ListView",
    "UpdateView",
]



from functools import reduce

from django.contrib.admin.utils import flatten
from django.core.exceptions import FieldDoesNotExist
from django.db.models.base import ModelBase
from django.forms.utils import pretty_name
from django.utils.html import format_html


class FieldService:
    JSON_SPLITTER = "__"

    pack = None

    def __init__(self, pack):
        self.pack = pack

    @classmethod
    def get_flatten_field_names(cls, field_names):
        if isinstance(field_names, dict):
            field_names = reduce(
                lambda accumulator, names: accumulator + flatten(names),
                field_names.values(),
                [],
            )
        else:
            field_names = flatten(field_names)

        return field_names

    @classmethod
    def get_botstrap_fields(cls, field_names, fields_data):
        def wrap(fields):
            fields = fields if isinstance(fields, (list, tuple)) else (fields,)
            cols = int(12 / len(fields))
            result = {field: fields_data.get(field) for field in fields}

            for value in result.values():
                value.update({"cols": cols})

            return result

        data = list(map(wrap, field_names))

        return data

    def _get_field_label(self, field_name):
        label = self.pack.default_labels.get(field_name)

        if label:
            return pretty_name(label)

        if "__str__" in field_name:
            label = self.pack.model._meta.verbose_name
            return pretty_name(label)

        names = field_name.split(self.JSON_SPLITTER)
        name = names.pop(0)

        try:
            field = self.pack.model._meta.get_field(name)
            label = field.verbose_name
        except FieldDoesNotExist:
            label = field_name

        return pretty_name(label)

    def _get_field_data(self, obj, large_field_name, field_name):
        if "__str__" == field_name:
            return {
                "value": obj,
                "label": self._get_field_label(large_field_name),
                "type": type(obj).__name__,
            }

        names = field_name.split(self.JSON_SPLITTER)
        name = names.pop(0)
        has_attr = name in obj if isinstance(obj, dict) else hasattr(obj, name)

        if not has_attr:
            raise AttributeError(
                "Does not exist attribute <{0}> for {1}".format(field_name, str(obj))
            )

        value = obj.get(name) if isinstance(obj, dict) else getattr(obj, name, None)
        value = value() if callable(value) else value
        is_file = hasattr(value, "field") and value.field.get_internal_type() in [
            "FileField",
            "ImageField",
        ]

        if is_file:
            value = {"name": value.name, "url": value.url} if bool(value) else None

        if len(names) and isinstance(value, dict):
            return self._get_field_data(
                value, large_field_name, self.JSON_SPLITTER.join(names)
            )

        try:
            field = obj._meta.get_field(name)
            internal_type = field.get_internal_type()

            if hasattr(field, "choices") and field.choices:
                value = dict(field.choices).get(value)
        except (FieldDoesNotExist, AttributeError):
            value = format_html(value)
            internal_type = type(value).__name__

        label = self._get_field_label(large_field_name)
        internal_type = internal_type.lower()

        return {"value": value, "label": label, "type": internal_type}

    def _get_data(self, instance_or_queryset, field_names):
        if isinstance(type(instance_or_queryset), ModelBase):
            instance_or_queryset = [instance_or_queryset]

        data = [
            (
                instance,
                {
                    field_name: self._get_field_data(instance, field_name, field_name)
                    for field_name in field_names
                },
            )
            for instance in instance_or_queryset
        ]

        return data

    def get_list_data(self, instance_or_queryset):
        return self._get_data(instance_or_queryset, self.pack.list_fields)

    def get_detail_data(self, instance_or_queryset):
        field_names = self.get_flatten_field_names(self.pack.detail_fields)
        data = self._get_data(instance_or_queryset, field_names) or {}

        if len(data):
            _, data = data[0]

        field_names = self.pack.detail_fields

        if not field_names:
            field_names = data.keys()

        fieldsets = (
            field_names.items()
            if isinstance(field_names, dict)
            else [("", field_names)]
        )

        bs_data = [
            {
                "title": title,
                "fieldset": self.get_botstrap_fields(
                    fieldset,
                    data,
                ),
            }
            for title, fieldset in fieldsets
        ]

        return data, bs_data



def get_base_view(ClassView, mixins, pack):
    class View(ClassView):
        def get_context_data(self, **kwargs):
            context = super().get_context_data(**kwargs)
            object = getattr(self, "object", None)
            config = {
                "title": self.model._meta.verbose_name_plural,
                "app_name": self.model._meta.app_label,
                "model_name": self.model._meta.model_name,
                "paths": self.pack.get_paths(object),
            }
            pack_info = context.get("pack_info", {})
            pack_info.update(config)
            context["pack_info"] = pack_info

            return context

        def get_template_names(self):
            self.template_name = self.pack._template_names.get(self.name)

            return super().get_template_names()

    View.__bases__ = (*mixins, *View.__bases__)
    View.pack = pack
    View.model = pack.model
    View.queryset = pack.queryset

    return View



from enum import Enum


class PackViews(Enum):
    LIST = "list"
    CREATE = "create"
    UPDATE = "update"
    DETAIL = "detail"
    DELETE = "delete"


def set_enum():
    for enum in PackViews:
        enum.suffix = enum.value
        enum.kwarg = ""
        enum.param = ""

        if enum not in (PackViews.LIST, PackViews.CREATE):
            enum.kwarg = "{id}"
            enum.param = "<int:pk>"

    PackViews.LIST.suffix = ""


set_enum()
